import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.NavigableMap;
import java.util.StringTokenizer;
import java.util.TreeMap;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.TextOutputFormat;
import org.apache.hadoop.mapreduce.Job;


/**
 * COSC 416 Midterm Question 5 Answer - Given a partkey, return the total number of times it was in an order, the total quantity ordered over all orders, and the average quantity ordered per order.
 *  
 */
public class Q5
{		
	/**
	 * Mapping function takes input of (LongWritable, Text) and produces output map of (NullWritable, Text).
	 * 
	 * This outputs all records to a NullWritable key.
	 
	 * The output key is a NullWritable so there is only one output key and only one reducer.
	 * The Text value is the entire tab-separate record that also contains the score.
	 */
    public static class Map extends MapReduceBase implements Mapper<LongWritable, Text, IntWritable, IntWritable>
    {        
    	// Stores the desired record id being searched for
    	private static int targetId;
    	
    	/**
    	 * 	Get game id parameter passed in job configuration object
    	 */
    	public void configure(JobConf job) 
    	{    	
    		targetId = Integer.parseInt(job.get("val"));
    	}
    	
        public void map(LongWritable key, Text value, OutputCollector<IntWritable, IntWritable> output, Reporter reporter) throws IOException
        {	// Output record
            String line = value.toString();
            StringTokenizer tokenizer = new StringTokenizer(line, "\t");
            tokenizer.nextToken();										// Order key
            String partIdStr = tokenizer.nextToken();					// Part key            
            tokenizer.nextToken();										// Supplier key
            tokenizer.nextToken();										// Linenumber
            String qty = tokenizer.nextToken();							// Quantity
            int quantity = Integer.parseInt(qty);
            int pkey =  Integer.parseInt(partIdStr);
            
            // Only keep target partkey
            if (targetId == pkey)
            	output.collect(new IntWritable(pkey), new IntWritable(quantity));
        }               
    }

    /**
     * Reduce function takes input of (IntWritable, IntWritable) and produces output map of (IntWritable, Text).
     *     
     */
    public static class Reduce extends MapReduceBase implements Reducer<IntWritable, IntWritable, IntWritable, Text>
    {           
        public void reduce(IntWritable key, Iterator<IntWritable> values, OutputCollector<IntWritable, Text> output, Reporter reporter) throws IOException
        {         
        	// Go through all the records and put them into TreeMap while only keeping the top N
        	long count = 0;
        	long sum = 0;
            while (values.hasNext())
            {  
                IntWritable val = values.next();
                count++;
                sum += val.get();                
            }       

            // Write out single output record            
            output.collect(key, new Text(""+key+"\t"+count+"\t"+sum+"\t"+(1.0*sum/count)));                                      
        }
    }

    /**
     * Main method to submit the job to Hadoop cluster after configurating job parameters.
     * @param args
     * 		arg[0] is record id wanted, arg[1] is input path, arg[2] is output path.  Defaults are used if nothing is specified.
     * @throws Exception
     * 		if any error occurs
     */
    public static void main(String[] args) throws Exception
    {
        JobConf conf = new JobConf(Q5.class);						// Name of class
        conf.setJobName("q5");										// Name of job when executing with Hadoop

        conf.setOutputKeyClass(IntWritable.class);					// Type of output key
        conf.setOutputValueClass(IntWritable.class);				// Type of output value
        
        conf.setMapperClass(Map.class);								// Class that implements mapper        
        conf.setReducerClass(Reduce.class);							// Class that implements reducer

        conf.setInputFormat(TextInputFormat.class);					// Indicate input files are in text format
        conf.setOutputFormat(TextOutputFormat.class);				// Indicate output files are in text format

        // This following line is required when executing remotely
        conf.set("mapred.jar","q5.jar");							// Configure the JAR that stores the class.  Change q1.jar to the name of your JAR file.
        
        Configuration conf2 = new Configuration();
        
        // Setup default input and output directories if not provided as arguments
        String outputPath = "/user/rlawrenc/416/midterm/";		// Default output path              
        String inputPath = "/user/rlawrenc/416/tpch/lineitem/lineitem.tbl";	// Default input path
        int recordId = 25;											// Default record id
        
        // Set input and output to command arguments if available
        if ( args.length > 0 ) 
        	recordId = Integer.parseInt(args[0]);
        if ( args.length > 1 ) 
        	inputPath = args[1];					
        if ( args.length > 2 ) 
        	outputPath = args[2];

        // Set a key-value pair with key val to indicate the record id that we need
        conf.set("val", ""+recordId);
        
        // Delete output path if it already exists as otherwise get an error if try to overwrite
        Path pt = new Path(outputPath);
        FileSystem fs = FileSystem.get(conf2);
        fs.delete(pt, true);		// Remove previous output directory       
        
        FileInputFormat.setInputPaths(conf, new Path(inputPath));
        FileOutputFormat.setOutputPath(conf, pt);
        
        Job job = new Job(conf, "q5");
        
        // Execute job
        job.waitForCompletion(true); 

        // Show output by iterating through all file parts (one part per reducer) and output lines
        FileStatus[] files = fs.listStatus(pt);
        int recordCount = 0;
        
        for (FileStatus file : files)
        {
            System.out.println("File: " + file.getPath());
            if ( file.isDir() || !file.getPath().toString().contains("part") ) continue;
            FSDataInputStream is = fs.open(file.getPath());
            BufferedReader br = new BufferedReader(new InputStreamReader(is));

            String s;
            while ((s = br.readLine()) != null)
            {
                System.out.println(s);
                recordCount++;
            }

            br.close();
        }
        fs.close();
        System.out.println("Total records printed: "+recordCount);
    }
}
